/* swpower_3_responses_V4.sas
* date: Sep 8th, 2021;
* Authors' name: John Preisser & Ying Zhang;
* ref: "Sample size determination for GEE analysis of stepped wedge cluster randomized trials";
* what: compute GEE power for binary/Count/Continuous outcomes and complete/incomplete SW-CRT 
  based on section 3.2 of Li, Turner and Preisser (2018), Humming's paper and section 4.1 in Rochon's paper ;                              
 
Input arguments:

*Designpattern: the macro allows the user to specify the design pattern matrix of a cluster trial: 0-control periods
		1- intervention periods, 2-  periods without data collection


* Cluster sizes and cluster numbers: The same cluster-period size(Ni) pattern per cluster, how many clusters are in a sequence(m).

* There are four correlation matrix structures to choose, 
   1- nested exchangeable/ 2- exponential decay , (for cross-sectional study)
   3 - block exchangeable/ 4 - proportional decay  (for cohort study)

* Two ways to model periods effect:
	1: In the model with categorical periods effect, there are T categorical periods effects.
	2: In the model with linear period effects, there are 2 parameters, one intercept and one slope of periods effect.

* Two intervention effect models:
  1: average intervention effects model/ 2: incremental intervention effects model;
  
* Three types of responses: 
* We included the fast GEE power calculation of three types of responses(binary, count, continuous) and different link functions.
* Default link function is the canonical link and the link option could be left blank in this case.
* The available link functions of binary outcomes are logit, log, and linear functions. Those for continuous and count responses are linear and log functions.
* There should be some warning messages if the estimated means are out of the responses' range.

* The overdispersion parameter is also specified in the model as phi, according to Rochon's paper.  


* Outputs:
	number of T timepoints (periods), S steps (sequences), m clusters per step, b periods before intervention at the first sequence, c periods between intervention across steps, 



* Section 3.2 of Li et al. (2018) describes var(d ^) and the Rochon's method for determining power.;

* Update: add the Frechet bound check for binary outcomes

*       alpha0 = ICC within period;(for exponential decay and proportional decay)
*       r0 = correlation decay rate over time;(for exponential decay and proportional decay)
*       alpha1 = ICC within period;(for block exchangeable and nested exchangeable)
*       alpha2 = ICC between period;(for block exchangeable and nested exchangeable)
*       alpha3 = ICC within person,  only used in the block exchangeable structure;
*       beta_period_effects = the parameters of periods effect  ;
*		max_intervention_period = the number of maximum intervention periods of incremental intervention effect model
*		completeness_matrix= a matrix with S x T  of 1’s and 0’s, with 0’s representing cluster-periods without data collection 
* 							and 1's standing for cluster-periods data collection 
*
* May 10: change the probnorm() function to probt() in power calculation*
* June 1: adapt the SAS macro into varying cluster in each sqeuence and open cohort study(varying correlation matrix for each sequence) 
* June 2nd: Test the SAS macro to parallel , SW-CRT, and CE CRTs, all matiching the results with Dr.Preisser's slides on 2020 ASA CT conference.
* June 14th: Extend the SAS macro with two types of degree of freedom using t-test;
* Sep 08th: Change the SAS macro name to CRTFASTGEEPWR;
*/


options ls=132;


%macro CRTFASTGEEPWR(
			alpha=, /*significance level, default 0.05*/
			m=,  /* a vector of the number of clusters in each sequence*/
			CP_size_matrix=, /*  number of cluster-period sizes(a matrix vector with dimension S*T), 
			       it is allowed to vary within a cluster across time, and clusters*/    
			corr_type=, /* correlation structure,1- nested exchangeable/ 2- exponential decay /
			                               3 - block exchangeable/ 4 - proportional decay  (for cohort study)*/
			intervention_effect_type= , /*the intervention effect model type: 1- average intervention effects model, 2- incremental intervention effects model */                           
			period_effect_type=,  /*the period effect model type: 1- categorical time period model, 2- continuous time period model*/ 
			alpha0=,  /* the within-period corr in exponential decay  and proportional decay*/
			r0=,       /* correlation decay rate over time in exponential decay  and proportional decay*/
			alpha1= , /* the within-period corr in nested exchangeable and block exchangeable*/
			alpha2= , /* the inter-period corr in nested exchangeable and block exchangeable*/
			alpha3= , /* the within-subject corr in block exchangeable*/
			delta= , /*the parameter of intervention,different meanings under differnt intervention effect model*/
			beta_period_effects = ,/* the parameters for period effects: a T*1 vector for categorical period effect/ a 2*1 vector for continuous period effect */
			max_intervention_period = , /* The number of maximum intervention periods of incremental intervention effect model*/
			DIST = , /* the distribution of responses: binary, poisson, normal  */
			link = , /* the available link function for GEE: logit, log, identity  */
			phi = , /*the dispersion parameter*/
			DesignPattern = ,/*  a matrix with S x T  of 0, 1 and 2, with 0 representing control periods, 1 standing for intervention periods, 2 presenting periods without data collection.*/ 
			df_choice =1  /*optional, 1 means df = I-p; 2 means df = I-2*/
			);
				
	proc IML ;
	
		
		
/**************************************************************************/		    
        /***set SAS modules for all correlation structure, refering Ying's notes***/
/********************************************************************/
         start nested_exchangeable(alpha1,alpha2,T, Ni); /*the SAS module for the nested exchangable corr, type=1*/
              	  
		  /*calculate the correlation matrix Ri*/ 
		    blockR = J(Ni[1],Ni[1],1);
			do j1=2 to T;
			    blockR = block(blockR, J(Ni[j1],Ni[j1],1));
			end;
			*print blockR;
			R = alpha2*J(sum(Ni),sum(Ni),1) + (alpha1- alpha2)*blockR + (1-alpha1)*I(sum(Ni));
			*print R;
			
			return(R) ;
         finish;
         /********************************************************************/
 
          start exp_decay(alpha0,r, T,Ni);  /*the SAS module for the exponential decay corr, type=2*/  	
		   A = J(T,T,0);/*obtain the AR(1) corr matrix A */
		   do j1=1 to T;
			  do j2 = j1 to T;
			    A[j1,j2] = r##abs(j2-j1);
			    A[j2,j1] = A[j1,j2];
			  end;
		   end;
			
			*print A;
			
		    R_sub = J(1,sum(Ni),0); /*create a row vector with 0, appending correlation matrix by each period*/
			do j1=1 to T;
			 Rj1 = J(Ni[j1],1,0);/*create a column vector with 0, appending correlation matrix of the j1th and j2th period*/
			  do j2 = 1 to T;
		         Rj1j2 = A[j1,j2]*J(Ni[j1],Ni[j2],1);
		         *print Rij;
		         Rj1 = Rj1||Rj1j2;
		         *print Ri;
				end;
			  trueRj1 = Rj1[,2:ncol(Rj1)];/*delete the former column vector with 0*/
			  R_sub = R_sub//trueRj1;
			end;
			
			
			trueR_sub = R_sub[2:nrow(R_sub),];/*delete the former row vector with 0*/
		    *print trueR_sub;
		    
		    
			R = alpha0*trueR_sub + (1-alpha0)*I(sum(Ni));
			/*print R;*/

			return (R);
           finish;
         /********************************************************************/
	       start block_exchange_R_inv(alpha1, alpha2, alpha3,T,N ); /*the SAS module for the block exchangable corr, type=3*/  
		
				/* eigenvalues;*/
				
				lambda1 = 1 - alpha1 + alpha2 - alpha3;
				lambda2 = 1 - alpha1 - (T-1)#(alpha2 - alpha3);
				lambda3 = 1 + (N-1)#(alpha1 - alpha2) - alpha3;
				lambda4 = 1 + (N-1)#alpha1 + (T-1)#(N-1)#alpha2 + (T-1)#alpha3;
				
				/* check that all lambdas are positive;
				print lambda1 lambda2 lambda3 lambda4;
				* compute inverse of R;
				also in Fan li paper, section 2 */
				aa = 1/lambda1;
				bb = (alpha3-alpha2)/(lambda1#lambda2);
				cc = (alpha1-alpha2)/(lambda1#lambda3);
				dd1 = (alpha3-alpha2)#(alpha1-alpha2)/(lambda1#lambda2#lambda3);
				dd2 = (alpha3#alpha1 - alpha2)/(lambda2#lambda3#lambda4);
				TN = T#N;
				R = (1-alpha1+alpha2-alpha3)#I(T#N) + (alpha3-alpha2)#J(T,T,1)@I(N) + (alpha1-alpha2)#I(T)@J(N,N,1)+alpha2*J(T#N,T#N,1);
				*Rinv = aa#I(TN) + bb#J(T,T,1)@I(N) - cc#I(T)@J(N,N,1) + (dd1 + dd2)#J(TN,TN,1);	/*the formula is in Fan's paper*/
				*print Rinv;
		        return( R );  /* return the inverse of Ri */
	        finish;
	        
        /********************************************************************/
         start proportional_decay(alpha0,rho, T,N);  /*the SAS module for the proportional decay corr, type=2*/  	
		   
		   A = J(T,T,0);/*obtain the AR(1) corr matrix A */
		   do j1=1 to T;
			  do j2 = j1 to T;
			    A[j1,j2] = rho##abs(j2-j1);
			    A[j2,j1] = A[j1,j2];
			  end;
		   end;
		   
		  B = alpha0*J(N,N,1) + (1-alpha0)*I(N);
		  R = (J(T,T,1)@B )#(A@J(N,N,1)) ;
		  *print Ri;
		  return (R); 

         finish;
         
         /*SAS module to chech frechet bound */
        start check_frechet(mu,cov);
		    dim1 = ncol(cov);
		      Do j = 1 to dim1-1;
		      	Do k = j+1 to dim1;
		      	flag = 0;
		      	  if cov[j,k] >=min(sqrt(mu[j]*(1-mu[k])/(mu[k]*(1-mu[j]))),sqrt(mu[k]*(1-mu[j])/(mu[j]*(1-mu[k]))))then do; flag =2; /*print j k flag;*/ return(flag); leave;end;
		      	  else if cov[j,k] <=max(-sqrt(mu[j]*mu[k]/((1-mu[k])*(1-mu[j]))),-sqrt(((1-mu[k])*(1-mu[j]))/(mu[j]*mu[k]))) then do; flag =1; /*print j k flag;*/ return(flag); leave; end;
		      	 end;
		      	end;
		      
		      return(flag); 
		finish; 
        
        
        
/**************************************************************************/		    
        /***Generate basic cluster randomized study elements***/
/********************************************************************/

		DesignPattern = &DesignPattern;
		S = nrow(DesignPattern);
		T = ncol(DesignPattern);
		*print S T;
		ref_cluster_period_matrix = &CP_size_matrix;
	    ind_CP_size_missing= (min(ref_cluster_period_matrix) =0);/*indicator of the incomplete CRT*/
	    
	    if ind_CP_size_missing=1  then do;
	    	loc_no_obs = loc(ref_cluster_period_matrix=0);
	    	ref_cluster_period_matrix[loc_no_obs] = 1;/*assuming there is at least one subjects in the reference CRT for incomplete CRTs*/
           %if &corr_type =3 |  &corr_type =4 %then %do;
        		ref_cluster_period_matrix[loc_no_obs] = &CP_size_matrix[1,1];/*assuming the consistent cohort size for incomplete closted-cohort CRT*/
 		   %end;
        end;

		/*Obtain the marignal model parameters from the inputs*/
       	delta = &delta; /* log odds ratio of intervention*/
		beta= &beta_period_effects; /* periods effect parameter*/		*print beta; *print delta;
		theta = beta//delta;/* set the total parameters theta;*/
		length_beta = nrow(beta);
		length_theta = nrow(theta);		
		
	   
/***************************decide the period effect in the design & consistency check 1***********************************************/		
		%if &period_effect_type = 1 %then %do; /*Choose the categorical period effect model, there are T period effects */
	        time_effect = I(T);	

	    %end;
	    %if &period_effect_type = 2 %then %do; /*Choose the continuous period effect model, a intercept and a slope for time. */
	        time_effect = J(T,1,1)||(0:(T-1))`; 
	    %end;	
	    ncol_tf = ncol(time_effect);
	    *print ncol_tf length_beta theta;
	    /*Two consistency checks about the number of parameters and chosen period effects model*/
	    if length_beta ^= ncol_tf  then
	    	Print "ERROR: The length of periods effect parameter does no fit with the periods effect model";
		if (T+1) ^= length_theta &   &&period_effect_type=1 then 
	    	Print "ERROR:The number of period T is not the same with the length of categorical periods effect";


/**************************************************************************/		    
        /***Calculate ZWZ by sequence and completeness pattern***/
/********************************************************************/


	    completeness_matrix = J(S,T,1);/*Generate the completeness matrix by the design pattern matrix, with the same dim*/
	    ind_missing_DP_matrix = (max(&DesignPattern)=2);

	    if ind_missing_DP_matrix=1 then do;
	    	idx_missing = loc(DesignPattern=2);
	    	completeness_matrix[idx_missing] = 0;
	    /*One consistency check about the missing locations*/	
	    if ncol(loc_no_obs) ^= ncol(idx_missing) then Print "ERROR: The missing cluster periods are different from the CP_size_matrix and design pattern matrix";
	    
		end;

		zwzsum=J(length_theta,length_theta,0);      /* The dim of zwzsum depends on how to period effect*/   
					/*In categorical period effect model, there are T period effects, so the dim of zwzsum is (T+1 * T+1)  */
					/*In continuous period effect model, there are 2 period-effects, so the dim of zwzsum is (3* 3)  */
		
		design_pattern = J(S,T,0);
		totaln = 0;
	    do i =1 to S;     /*generate the zwzsum by adding zwzi by each sequence*/
	   		Ni = ref_cluster_period_matrix[i,];
	   		totaln = totaln + sum(&m[i,]*&CP_size_matrix[i,]); /*The total number of subjects are based on the real clutster period size matrix*/
 /***********generate correlation matrix by selected type*********************************************************/

         	%if &corr_type = 1 %then %do;
         	   alpha1 = &alpha1;
         	   alpha2 = &alpha2;
		       R =   nested_exchangeable(alpha1,alpha2,T, Ni);
		       %let ctype = nested exchangeable correlation structure;
		       %let  c_param = (alpha1,alpha2):(&alpha1, &alpha2);
		    %end;
		    
		    %if &corr_type = 2 %then %do;
		       alpha0 = &alpha0;
         	   r_0 = &r0 ;
		       R =  exp_decay(alpha0,r_0, T,Ni);
			%let ctype = exponential decay correlation structure;
			%let c_param = (alpha0,r0):(&alpha0, &r0);
		    %end;

        %if &corr_type = 3 %then %do;
		       N = Ni[1,1]; /*Closed_cohort: assume same cluster-period size for each cluster*/
         	   alpha1 = &alpha1;
         	   alpha2 = &alpha2;
         	   alpha3 = &alpha3;
		       R =   block_exchange_R_inv(alpha1, alpha2, alpha3,T,max(Ni));
		       %let ctype = block exchangeable correlation structure;
		       %let c_param = (alpha1,alpha2,alpha3):(&alpha1, &alpha2, &alpha3);
		    %end;
		    
		    %if &corr_type = 4 %then %do;
		       N = Ni[1,1]; /*Closed_cohort: assume same cluster-period size for each cluster*/
		       alpha0 = &alpha0;
         	   r0 = &r0 ;
		       R =   proportional_decay(alpha0, r0,T,max(Ni));
		       %let ctype = proportional decay correlation structure;
		       %let c_param = (alpha0,r0):(&alpha0, &r0);
		    %end;
		   *print R ; 


		/*generate the design matrix by the intervention effects*/
            %if &intervention_effect_type = 1 %then %do; /*Choose the average intervention effects model */
                 Xi = DesignPattern[i,] ;      /* Xi is trt sequence received by all clusters in sequence i;*/        
	        %end;
	        
	        %if &intervention_effect_type = 2 %then %do; /*Choose the incremental intervention effects model*/
	            locs = loc(DesignPattern[i,]=1);
	            Xi = J(1,(locs[1]-1),0)||((1:(T-locs[1]+1))/&max_intervention_period) ;      /* Xi is trt sequence received by all clusters in sequence i;*/        
		    %end;
			Xi = Xi`;
			completeness_seq_i = completeness_matrix[i,];/*get the completeness pattern by each sequence*/
			/*expand the completeness sequence into matrix by the cluster period in the sth step*/
			block_completeness_matrix = completeness_seq_i[1]*I(Ni[1]); /*repeat the design matrix row by cluster-period size*/
			do j=2 to T;
					block_completeness_matrix = block(block_completeness_matrix, completeness_seq_i[j]*I(Ni[j]));/*for the block matrix, the row with all zeros indicating periods without observations*/
			end;  

			Ki= block_completeness_matrix[loc(block_completeness_matrix[+,]=1),]; /*delete the rows with all zeros */

        
	        *print Xi design_pattern;
	        design_pattern[i,] = Xi`;  
	        design_matrix_i = (time_effect||Xi); /*append the time effect with intervention effect*/
	       	*print  design_matrix_i;      
	        Zi = design_matrix_i[1,]@J(Ni[1],1,1);  /* initialize Zi with the first period */
	        do j = 2 to T;
	  	    	Zi = Zi//(design_matrix_i[j,]@J(Ni[j],1,1)); /* generate Zi by varying cluster-period size Ni */
	        end;

		    /*obtain the mean and variance matrix by responses types and link functions*/
		    gi = Zi*theta;
		    *print Zi gi;
		    Dist = &Dist;
		    
		    %if &Dist = "binary" %then %do; /*for binary outcomes*/
				/*check the Frechet bound for binary outcomes*/
				Zi_star = Ki*Zi;
		    	gi_star = Zi_star*theta;
				mui_star = exp(gi_star)/(1 + exp(gi_star)); /*logit link as the default link function*/
	       		Ri_star = Ki*R*Ki`;
				flag = check_frechet(mui_star,Ri_star);/*The frechet boundary is examinated based on the true incomplete average mean vector mui_star*/
				if flag>0 then do; print "ERROR: Correlation out of Frechet bound";Leave;end;
				mui = exp(gi)/(1 + exp(gi)); /*logit link as the default link function*/
				*print mui;
				Deltai = diag(mui#(1-mui));
				Link = "logit";/*canonical link*/
		    	%if &link = "log" %then %do;
		    		Link = &link;
			    	mui = exp(gi); /*outcome mean under log link */
			    	/*internal consistency checks for binary outcomes*/
			    	if max(mui) >=1 then 
			    		Print "ERROR: The log link of binary outcome does not work as mean is larger than 1";
			    	Deltai = diag(mui);
		    	%end;
		    	%if &Link = "identity" %then %do;
		    		Link = &link;
			    	mui = gi; /*outcome mean under identity link */
			    	/*internal consistency checks for binary outcomes*/
			    	if max(mui) >=1 | min(mui) <=0  then 
			    		print "ERROR: The identity link of binary outcome does not work as mean is larger than 1 or smaller than 0";
			    	Deltai = I(nrow(mui));
		    	%end;
		    	Ai_sqrt =  diag(sqrt(mui#(1-mui)));/*Ai_sqrt is the square root of variance matrix Ai, which is sqrt(mu*(1-mu)) in the binary outcomes */
                *print mui Ai_sqrt Deltai ;
		    %end;
		    
		    
		    
		    %if &Dist = "poisson" %then %do;/*for count outcomes*/
		    	mui = exp(gi); /*log link as the default link function*/
				Deltai = diag(mui);
				Link = "log";/*canonical link*/
				%if &Link = "identity" %then %do;
				    Link = &link;
			    	mui = gi; /*outcome mean under identity link */
			    	/*internal consistency checks for count outcomes*/
			    	if min(mui) <=0  then 
			    		print "ERROR: The identity link of count outcome does not work because the mean is negative";
			    	Deltai = I(nrow(mui));
		    	%end;
		    	Ai_sqrt =  diag(sqrt(mui));/*Ai_sqrt is the square root of variance matrix Ai, which is sqrt(mu) in the count outcomes  */
		    
		    %end;
		    
		    
		    %if &Dist = "normal" %then %do;
		    	mui = gi; /*identity link as the default link function*/
				Deltai = I(nrow(mui));
				Link = "identity";/*canonical link*/
				%if &Link = "log" %then %do;
				    Link = &link;
			    	mui = exp(gi); /*outcome mean under log link */
			    	/*internal consistency checks for normal outcomes*/
			    	Deltai = diag(mui);
		    	%end;
		    	Ai_sqrt =  I(nrow(mui));/*Ai_sqrt is the square root of variance matrix Ai, which is sqrt(mu) in the count outcomes  */
		
		    %end;


			*print mui;
		    Zi_star = Ki*Zi;
		    Ai_sqrt_star = Ki*Ai_sqrt*Ki`;
		    Deltai_star = Ki*Deltai*Ki`;
	        Wi_star = Deltai_star*inv(Ai_sqrt_star*Ki*R*Ki`*Ai_sqrt_star)*Deltai_star;
	        *print Wi;
	        zwzi = t(Zi_star)*Wi_star*Zi_star;
	        zwzsum = zwzsum + zwzi*&m[i,];/* specify the number of clusters per sequence */
	        *print zwzsum;
		
		end;
		*print i block_completeness_matrix  Ki R;
		
	    *print design_pattern ;
	    *print completeness_matrix; 
		*real_intervention_effect = design_pattern#completeness_matrix;
		/*////print real_intervention_effect;*/
		phi = &phi; /*set the dispersion parameters*/
		modvar = phi*inv(zwzsum);/*calculate the model based variance*/
		*print zwzsum;
		vardel = modvar[length_theta, length_theta];  /*pick the variance of intervention effect*/

		*print vardel;
				
		alpha=&alpha; /*get value from macro*/
		twotailprob=(1-alpha/2);
		ztwotail=probit(twotailprob);
		*print ztwotail;
		
		*n=vardel*(ztwotail+zpower)**2/(del0**2); /* for 80% power this is phi*3.92/(m*diff*diff) */
		stddel = abs(delta)/sqrt(vardel);
		*print stddel;
		zpower_value = stddel - ztwotail;
		zpower = probnorm(zpower_value);
		*print zpower;
		
		clusters = sum(&m);
		* p is the number of parameters estimated;
		df = clusters - length_theta;  /*df impacted by choice of modeling time effect */
		%if &df_choice=2 %then %do; df = clusters-2;%end;
		
		ttwotail = tinv(1-alpha/2, df); 
		tpower_value = stddel - ttwotail;
		tpower = probt(tpower_value,df);
		*print tpower;
		dist  = &dist;
		corr_type = &Corr_type;
        /********************************************************************/
	    %let name_dist = %qsysfunc(compress(&DIST,%str(%")));
	    title "The fast GEE power of &name_dist outcomes with &ctype and &c_param";
	    %if &intervention_effect_type=1 %then %do;
		title2 "Under average intervention effects model and delta = &delta";

		%end;
		%if &intervention_effect_type=2 %then %do;
		title2 "Under incremental intervention effects model and delta = &delta";
		%end;
		print T S clusters df theta totaln dist link stddel zpower tpower ;		

		quit;  

%mend; 
